use bevy::{
    prelude::*,
    sprite::{MaterialMesh2dBundle, Mesh2dHandle},
    utils::{Duration, HashSet},
};
use bevy_rand::prelude::*;
use bevy_rapier2d::prelude::*;
use rand::Rng;

use crate::{
    game_state::GameTimer,
    player::{LifeChangeEvent, Player},
    scene::SceneObj,
    GameState, GameplaySet, METER,
};

#[derive(Component)]
struct SpawnTimer(pub Timer);

#[derive(Component)]
struct DroppedCrate;

#[derive(Debug, Event)]
struct CrateDropEvent {
    pos: f32,
}

pub fn spawner_plugin(app: &mut App) {
    app.add_plugins(EntropyPlugin::<WyRand>::default())
        .add_event::<CrateDropEvent>()
        .add_event::<CrateCollision>()
        .add_systems(OnEnter(GameState::InGame), add_timer.in_set(GameplaySet))
        .add_systems(OnExit(GameState::InGame), cleanup.in_set(GameplaySet))
        .add_systems(
            Update,
            (
                drop_crates,
                do_drop,
                crate_collisions,
                delete_on_env_coll,
                player_coll,
                update_timer_for_difficulty,
            )
                .in_set(GameplaySet),
        );
}

fn cleanup(
    timer: Query<Entity, With<SpawnTimer>>,
    crates: Query<Entity, With<DroppedCrate>>,
    mut commands: Commands,
) {
    commands.entity(timer.single()).despawn();
    for e in &crates {
        commands.entity(e).despawn();
    }
}

fn add_timer(mut commands: Commands) {
    commands.spawn(SpawnTimer(Timer::from_seconds(0.2, TimerMode::Repeating)));
}

fn update_timer_for_difficulty(
    mut spawn_timer: Query<&mut SpawnTimer>,
    mut game_timer: Query<&mut GameTimer>,
    time: Res<Time>,
) {
    let game_time = game_timer
        .single_mut()
        .game_time
        .tick(time.delta())
        .elapsed_secs() as i32;

    println!("time: {game_time}s");

    spawn_timer
        .single_mut()
        .0
        .set_duration(Duration::from_millis(match game_time {
            0..=9 => 2000,
            10..=19 => 1000,
            20..=29 => 500,
            30..=39 => 400,
            40..=49 => 350,
            50..=59 => 300,
            60..=69 => 250,
            70..=79 => 200,
            80..=89 => 175,
            90..=99 => 150,
            100..=109 => 125,
            110..=119 => 100,
            120..=129 => 90,
            130..=139 => 80,
            140..=149 => 70,
            150..=159 => 60,
            160..=169 => 50,
            170..=179 => 25,
            180.. => 10,
            ..=-1 => unreachable!(),
        }));
}

fn drop_crates(
    mut ev_drop: EventWriter<CrateDropEvent>,
    mut rng: ResMut<GlobalEntropy<WyRand>>,
    mut spawn_timer: Query<&mut SpawnTimer>,
    time: Res<Time>,
) {
    let mut timer = spawn_timer.single_mut();

    if timer.0.tick(time.delta()).finished() {
        ev_drop.send(CrateDropEvent {
            pos: rng.gen_range(-7f32..7f32),
        });
    }
}

fn do_drop(
    mut ev_drop: EventReader<CrateDropEvent>,
    mut commands: Commands,
    mut meshes: ResMut<Assets<Mesh>>,
    mut materials: ResMut<Assets<ColorMaterial>>,
) {
    let rect = Mesh2dHandle(meshes.add(Rectangle::new(1. * METER, 1. * METER)));
    for ev in ev_drop.read() {
        println!("ev: {ev:?}");
        commands
            .spawn(DroppedCrate)
            .insert(MaterialMesh2dBundle {
                mesh: rect.clone(),
                material: materials.add(Color::rgb(0.9, 0.8, 0.2)),
                transform: Transform::from_xyz(ev.pos * METER, 12. * METER, 0.),
                ..default()
            })
            .insert(ActiveEvents::COLLISION_EVENTS)
            .insert((
                RigidBody::Dynamic,
                Collider::cuboid(0.5 * METER, 0.5 * METER),
            ));
    }
}

#[derive(Event)]
struct CrateCollision {
    coll_crate: Entity,
    with: CollisionType,
}

enum CollisionType {
    Player,
    Scene,
}

fn crate_collisions(
    mut player: Query<Entity, With<Player>>,
    mut ev_colls: EventWriter<CrateCollision>,
    drops: Query<Entity, With<DroppedCrate>>,
    scene_objs: Query<Entity, With<SceneObj>>,
    mut collision_events: EventReader<CollisionEvent>,
) {
    let p = player.single_mut();
    let crates = drops.iter().collect::<HashSet<_>>();
    let scene_objs = scene_objs.iter().collect::<HashSet<_>>();
    for collision_event in collision_events.read() {
        match collision_event {
            CollisionEvent::Started(e1, e2, _) if crates.contains(e1) || crates.contains(e2) => {
                let (coll_type, crate_) = if scene_objs.contains(e1) {
                    (CollisionType::Scene, e2)
                } else if scene_objs.contains(e2) {
                    (CollisionType::Scene, e1)
                } else if *e2 == p {
                    (CollisionType::Player, e1)
                } else if *e1 == p {
                    (CollisionType::Player, e2)
                } else if crates.contains(e1) && crates.contains(e2) {
                    continue;
                } else {
                    // there can't be anything else to collide with,
                    // only either scene objects or the player
                    eprintln!("e1: {e1:?}, e2: {e2:?}");
                    unreachable!()
                };
                ev_colls.send(CrateCollision {
                    coll_crate: *crate_,
                    with: coll_type,
                });
            }
            _ => {}
        }
    }
}

fn delete_on_env_coll(mut ev_colls: EventReader<CrateCollision>, mut commands: Commands) {
    for CrateCollision { coll_crate, with } in ev_colls.read() {
        if let CollisionType::Player = with {
            continue;
        }

        if let Some(mut e) = commands.get_entity(*coll_crate) {
            e.despawn();
        }
    }
}

fn player_coll(
    mut ev_colls: EventReader<CrateCollision>,
    mut ev_lives: EventWriter<LifeChangeEvent>,
    mut commands: Commands,
) {
    for CrateCollision { coll_crate, with } in ev_colls.read() {
        if let CollisionType::Scene = with {
            continue;
        }

        ev_lives.send(LifeChangeEvent::Lost);

        if let Some(mut e) = commands.get_entity(*coll_crate) {
            e.despawn();
        }
    }
}
