#![allow(clippy::type_complexity)]
use crate::main_menu::main_menu_plugin;
use bevy::prelude::*;
use bevy_rapier2d::prelude::*;
use drops::spawner_plugin;
use game_over_menu::{game_over_menu_plugin, GameOverMenuSet};
use game_state::state_and_ui_plugin;
use main_menu::MainMenuSet;
use player::{player_plugin, GameOverEvent};
use scene::scene_plugin;

const METER: f32 = 48.;

#[derive(SystemSet, Debug, Clone, PartialEq, Eq, Hash)]
struct GameplaySet;

#[derive(States, Debug, Clone, PartialEq, Eq, Hash)]
enum GameState {
    MainMenu,
    InGame,
    GameOver,
}

fn main() {
    let mut rapier_config = RapierConfiguration::new(METER * 2.5);
    rapier_config.timestep_mode = TimestepMode::Variable {
        max_dt: 1.0 / 60.0,
        time_scale: 1.0,
        substeps: 1,
    };
    App::new()
        .add_plugins(DefaultPlugins)
        // this does things with gravity
        .insert_resource(rapier_config)
        .insert_state(GameState::MainMenu)
        .add_plugins(RapierPhysicsPlugin::<NoUserData>::pixels_per_meter(METER))
        // .add_plugins(RapierDebugRenderPlugin::default())
        .add_event::<GameOverEvent>()
        .add_plugins((
            main_menu_plugin,
            spawner_plugin,
            state_and_ui_plugin,
            player_plugin,
            scene_plugin,
            game_over_menu_plugin,
        ))
        .add_systems(Startup, setup_cam)
        .add_systems(Update, handle_game_over.in_set(GameplaySet))
        .configure_sets(
            Update,
            (
                GameplaySet.run_if(in_state(GameState::InGame)),
                MainMenuSet.run_if(in_state(GameState::MainMenu)),
                GameOverMenuSet.run_if(in_state(GameState::GameOver)),
            ),
        )
        .run()
}

fn handle_game_over(
    mut ev_gameover: EventReader<GameOverEvent>,
    mut game_state: ResMut<NextState<GameState>>,
) {
    for _ in ev_gameover.read() {
        game_state.set(GameState::GameOver)
    }
}

fn setup_cam(mut commands: Commands) {
    commands.spawn(Camera2dBundle {
        transform: Transform::from_xyz(0., 4. * METER, 0.),
        ..default()
    });
}

mod drops;
mod game_over_menu;
mod game_state;
mod main_menu;
mod player;
mod scene;
