use bevy::{app::AppExit, prelude::*};

use crate::GameState;

#[derive(SystemSet, Debug, Clone, PartialEq, Eq, Hash)]
pub struct MainMenuSet;

pub fn main_menu_plugin(app: &mut App) {
    app.add_systems(
        OnEnter(GameState::MainMenu),
        setup_main_menu_ui.in_set(MainMenuSet),
    )
    .add_systems(Update, button_action.in_set(MainMenuSet))
    .add_systems(Update, button_color_system.in_set(MainMenuSet))
    .add_systems(OnExit(GameState::MainMenu), exit_menu.in_set(MainMenuSet));
}

#[derive(Component)]
struct MainMenu;

#[derive(Component)]
enum MenuButtonAction {
    Play,
    Quit,
}

fn setup_main_menu_ui(mut commands: Commands) {
    commands
        .spawn((
            NodeBundle {
                style: Style {
                    align_items: AlignItems::Center,
                    justify_content: JustifyContent::Center,

                    flex_direction: FlexDirection::Column,
                    width: Val::Percent(100.0),
                    height: Val::Percent(100.0),
                    ..default()
                },
                ..default()
            },
            MainMenu,
        ))
        .with_children(|parent| {
            parent.spawn(TextBundle::from_section(
                "Ev4der",
                TextStyle {
                    font_size: 144.,
                    ..default()
                },
            ));
            let btn_style = Style {
                width: Val::Px(300.0),
                height: Val::Px(80.0),
                margin: UiRect::all(Val::Px(20.0)),

                justify_content: JustifyContent::Center,
                align_items: AlignItems::Center,
                ..default()
            };
            let btn_txt_style = TextStyle {
                font_size: 64.,
                ..default()
            };
            parent
                .spawn((
                    ButtonBundle {
                        style: btn_style.clone(),
                        background_color: Color::rgb(0.3, 0.3, 0.3).into(),
                        ..default()
                    },
                    MenuButtonAction::Play,
                ))
                .with_children(|parent| {
                    parent.spawn(TextBundle::from_section("Play!", btn_txt_style.clone()));
                });
            parent
                .spawn((
                    ButtonBundle {
                        style: btn_style.clone(),
                        background_color: Color::rgb(0.3, 0.3, 0.3).into(),
                        ..default()
                    },
                    MenuButtonAction::Quit,
                ))
                .with_children(|parent| {
                    parent.spawn(TextBundle::from_section("Quit :(", btn_txt_style.clone()));
                });
        });
}

fn exit_menu(to_despawn: Query<Entity, With<MainMenu>>, mut commands: Commands) {
    for e in &to_despawn {
        commands.entity(e).despawn_recursive();
    }
}

pub fn button_color_system(
    mut interaction_query: Query<
        (&Interaction, &mut BackgroundColor),
        (Changed<Interaction>, With<Button>),
    >,
) {
    for (interaction, mut color) in &mut interaction_query {
        *color = match *interaction {
            Interaction::Pressed => Color::rgb(0.4, 0.4, 0.4).into(),
            Interaction::Hovered => Color::rgb(0.2, 0.2, 0.2).into(),
            Interaction::None => Color::rgb(0.3, 0.3, 0.3).into(),
        }
    }
}

fn button_action(
    interaction_query: Query<
        (&Interaction, &MenuButtonAction),
        (Changed<Interaction>, With<Button>),
    >,
    mut ev_exit: EventWriter<AppExit>,
    mut game_state: ResMut<NextState<GameState>>,
) {
    for (interaction, action) in &interaction_query {
        if *interaction == Interaction::Pressed {
            match action {
                MenuButtonAction::Quit => {
                    ev_exit.send(AppExit);
                }
                MenuButtonAction::Play => {
                    game_state.set(GameState::InGame);
                }
            }
        }
    }
}
