use bevy::{
    prelude::*,
    sprite::{MaterialMesh2dBundle, Mesh2dHandle},
    utils::HashSet,
};
use bevy_rapier2d::prelude::*;

use crate::{GameState, GameplaySet, METER};

#[derive(Component)]
pub struct Player {
    move_cooldown: Timer,
    grounded: bool,
    lives: u8,
}

#[allow(unused)]
#[derive(Event)]
pub enum LifeChangeEvent {
    Gained,
    Lost,
}

#[derive(Event)]
pub struct GameOverEvent;

const DEFAULT_LIVES: u8 = 5;

pub fn player_plugin(app: &mut App) {
    app.add_event::<LifeChangeEvent>()
        .add_systems(
            OnEnter(GameState::InGame),
            (add_player, add_lives_ui).in_set(GameplaySet),
        )
        .add_systems(
            OnExit(GameState::InGame),
            cleanup_player.in_set(GameplaySet),
        )
        .add_systems(
            Update,
            (move_player, player_ground_collision, update_lives).in_set(GameplaySet),
        );
}

fn add_player(
    mut commands: Commands,
    mut meshes: ResMut<Assets<Mesh>>,
    mut materials: ResMut<Assets<ColorMaterial>>,
) {
    let shape = Mesh2dHandle(meshes.add(Rectangle::new(0.8 * METER, 1.8 * METER)));
    commands
        .spawn(Player {
            move_cooldown: Timer::from_seconds(0.01, TimerMode::Repeating),
            // move_cooldown: Timer::from_seconds(0.01, TimerMode::Repeating),
            grounded: false,
            lives: DEFAULT_LIVES,
        })
        .insert(MaterialMesh2dBundle {
            mesh: shape,
            material: materials.add(Color::rgb(0.2, 0.9, 0.2)),
            transform: Transform::from_xyz(0., 0., 0.),
            ..default()
        })
        .insert((
            RigidBody::Dynamic,
            Collider::cuboid(0.4 * METER, 0.9 * METER),
            Velocity::default(),
        ))
        // .insert(ActiveCollisionTypes::default() | ActiveCollisionTypes::DYNAMIC_STATIC)
        .insert(ActiveEvents::COLLISION_EVENTS)
        .insert(KinematicCharacterController { ..default() });
}

fn cleanup_player(
    player: Query<Entity, With<Player>>,
    lives_ui: Query<Entity, With<LivesUi>>,
    mut commands: Commands,
) {
    commands.entity(player.single()).despawn();
    commands.entity(lives_ui.single()).despawn();
}

#[derive(Component)]
struct LivesUi;
fn add_lives_ui(mut commands: Commands) {
    commands.spawn((
        TextBundle::from_section(
            "<3 ".repeat(DEFAULT_LIVES.into()),
            TextStyle {
                font_size: 56.,
                ..default()
            },
        ),
        LivesUi,
    ));
}

fn update_lives(
    mut ev_lifechange: EventReader<LifeChangeEvent>,
    mut ev_gameover: EventWriter<GameOverEvent>,
    mut player: Query<&mut Player>,
    mut ui: Query<&mut Text, With<LivesUi>>,
) {
    let mut txt = ui.single_mut();
    let mut p = player.single_mut();

    for ev in ev_lifechange.read() {
        match ev {
            LifeChangeEvent::Lost => p.lives -= 1,
            LifeChangeEvent::Gained => p.lives += 1,
        }

        if p.lives == 0 {
            ev_gameover.send(GameOverEvent);
        } else {
            txt.sections[0].value = "<3 ".repeat(p.lives.into());
        }
    }
}

fn player_ground_collision(
    mut player: Query<(Entity, &mut Player)>,
    scene_objs: Query<Entity, With<super::scene::SceneObj>>,
    mut collision_events: EventReader<CollisionEvent>,
) {
    let (p, mut player) = player.single_mut();
    let scene_objs = scene_objs.iter().collect::<HashSet<_>>();

    for collision_event in collision_events.read() {
        match collision_event {
            CollisionEvent::Started(e1, e2, _) => {
                if (e1 == &p && scene_objs.contains(e2)) || (e2 == &p && scene_objs.contains(e1)) {
                    player.grounded = true;
                }
            }
            CollisionEvent::Stopped(e1, e2, _) => {
                if (e1 == &p && scene_objs.contains(e2)) || (e2 == &p && scene_objs.contains(e1)) {
                    player.grounded = false;
                }
            }
        }
    }
}

fn move_player(
    kb_input: Res<ButtonInput<KeyCode>>,
    mut query: Query<(
        &mut Velocity,
        &mut Player,
        &mut KinematicCharacterController,
    )>,
    time: Res<Time>,
) {
    let (mut vel, mut player, mut controller) = query.single_mut();
    if player.move_cooldown.tick(time.delta()).finished() {
        let mut moved = false;
        let mut mv = 0;

        if kb_input.pressed(KeyCode::KeyA) {
            moved = true;
            mv -= 1;
        }
        if kb_input.pressed(KeyCode::KeyD) {
            moved = true;
            mv += 1;
        }

        if kb_input.pressed(KeyCode::KeyW) && player.grounded {
            moved = true;
            *vel = Velocity::linear(Vec2::new(0., 9. * METER));
            // *vel = Velocity::linear(Vec2::new(0., 9.81 * METER));
        }

        if moved {
            controller.translation = Some(Vec2::new(mv as f32 * 6., 0.));
        }
    }
}
